/* Copyright 2017, 2018, 2021 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Search all trees using gentreeg for certain minimal domsets count.

   even
   n=28  2023443032 trees  max minimal_domsets=16384 count 3159

   n=30  2813481573
   n=30 14830871802 target


   Subdivision
   -----------

   n=28  2023443032 trees  max minimal_domsets=201310209 count 1
   cf pow 209523980.324
   :vm?OCGBB?p?KSBE?poK_BH?q_KkBK?rOKwBN?S?DC@Q?SoDO@T?T_D[@W?UODg@Z?^
   vpar_count_free(28)
   (12161.0^(1/27.0))^(2*28-1)

   2*29-1 == 57
   (12161.0^(1/27.0))^(2*29-1)
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_gentreeg_main(int argc, char**argv);

const int option_verbose = 0;
const int option_subdivide = 1;

unsigned long long  trees_total_domnum[64];

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))


/* Good up to n=29 at least, but somewhere later the actual count is bigger
   than this.

   n=27    751,065,460 trees
   :Z_`a`c`e`g`i`k_mnmpmrmtmvmx
   minimal_domsets maximum 12261 for 6,6

   n=29  5,469,566,585 trees
   :\_`a`c`e`g`i`k`m_oporotovoxoz
   minimal_domsets maximum 24385 for 6,7
 */
unsigned long long
small_minimal_domsets_count_max (int n)
{
  static const unsigned long long  table[] = { 0, 3ULL, 0, 2ULL };
  return ((n&1 ? 3ULL : 1ULL) << ((n>>1)-(n&1)))  /* [1, 3/2] * 2^floor(n/2) */
    - (table[n&3] << ((n-2)>>2))         /* [0, 3/2, 0, 2],n) * 2^floor(n/4) */
    + (n&1);    /* [0,1] */
}

int  target_n = 0;
unsigned long long  target;
int view_count = 0;
unsigned long long  tree_count = 0;
unsigned long long  max_value = 0;
int  max_count = 0;
DYNALLSTAT(char,max_g6,max_g6_sz);
double two_pow;

void
my_gentreeg_outproc (FILE *outfile, int *vpar, int n)
{
  if (target_n != n) {
    if (max_g6 && max_g6[0]) {
      printf("n=%d  %Lu trees  max minimal_domsets=%Lu count %d\n",
             target_n, tree_count, max_value, max_count);
      printf(" cf pow %.3f\n", two_pow);
      printf("%s\n", max_g6);
      /* nautyextra_view_g6(max_g6); */
    }

    target_n = n;
    target = small_minimal_domsets_count_max (n);
    max_value = 0;
    max_count = 0;
    DYNALLOC1(char,max_g6,max_g6_sz,1,"malloc");
    max_g6[0] = '\0';
    tree_count = 0;
    two_pow = pow(2, (double) n / 2);
    if (option_subdivide) {
      int actual_n = n+(n-1);
      two_pow = pow (12161.0, 1/27.0 * actual_n);
    }
  }
  tree_count++;

  if (option_subdivide) {
    static int new_vpar[256];
    memcpy(new_vpar, vpar, (n+1)*sizeof(*new_vpar));
    vpar = new_vpar;
    n = nautyextra_subdivide_vpar (vpar,n, 1);
  }

  if (option_verbose) {
    printf("tree %p  n=%d\n", vpar,n);
    nautyextra_print_vpar(vpar,n);
  }

  unsigned long long  count_minimal_domsets
    = nautyextra_minimal_domsets_count_vpar(vpar,n);
  if (max_value < count_minimal_domsets) {
    max_value = count_minimal_domsets;
    max_count = 0;

    const char *str = nautyextra_vpar_to_s6(vpar,n);
    DYNALLOC1(char,max_g6,max_g6_sz,strlen(str)+1,"malloc");
    strcpy (max_g6, str);
#if 0
    printf("n=%d new high max_value %Lu  (at tree %Lu)\n",
           n, max_count, tree_count);
#endif
  }
  if (max_value == count_minimal_domsets) {
    max_count++;
  }

  if (n <= 119 && (double) count_minimal_domsets > two_pow) {
    if (view_count++ < 10) {
      printf("more than 2pow %Lu\n", (unsigned long long) count_minimal_domsets);
      nautyextra_print_vpar(vpar,n);
      printf("%s\n", nautyextra_vpar_to_s6(vpar,n));
      nautyextra_view_vpar(vpar,n);
    }
  }

  if (NAUTYEXTRA_PROGRESS)
    printf("n=%d %Lu max %Lu count %d\r", n, tree_count, max_value, max_count);

#if 0
  if (count_minimal_domsets == target && view_count++ < 25) {
    printf("vpar "); nautyextra_print_vpar(vpar,n);
  }
#endif
}

int
main (void)
{
  setbuf(stdout,NULL);
  nautyextra_view_synchronous = 1;
  nautyextra_want_progress = 1;

  if (0) {
    int vpar[100];
    int n = 4;
    nautyextra_make_star_vpar(vpar,n);
    nautyextra_print_vpar(vpar,n);
    nautyextra_view_vpar(vpar,n);
    n = nautyextra_subdivide_vpar (vpar,n, 3);
    nautyextra_print_vpar(vpar,n);
    nautyextra_view_vpar(vpar,n);
  }

  {
    char *str = ":Z_`abc`e`g`i`k_mnopmrmtmvmx";
    int loops;
    int vpar[100];
    SG_DECL(sg);
    SG_INIT(sg);
    stringtosparsegraph(str,&sg,&loops);
    int n = sg.nv;
    int root;
    for (root = 0; root < n; root++) {
      nautyextra_sg_to_vpar(&sg, vpar, root);
      /* nautyextra_view_vpar(vpar,n); */

      unsigned long long  count = nautyextra_minimal_domsets_count_vpar(vpar,n);
      printf("root=%d  count %Lu\n", root, count);
    }

    SG_FREE(sg);
  }

  {
    /* A306692 maximum num minimal dominating sets */
    static const unsigned long long  wants[]
      = { 2,2,4,4,8,9,16,19,32,41,64,85,128,177,256,361,512,737,1024,1489,2048,3009 };
    int i;
    for (i=1; i<sizeof(wants)/sizeof(wants[0]); i+=2) {
      int n = i+2;
      unsigned long long  want = wants[n-2];
      unsigned long long  got = small_minimal_domsets_count_max(n);
      if (got != want) {
        printf("n=%d wrong got %Lu want %Lu\n", n, got, want);
      }
    }
    printf("\n");
  }
  {
    int n;
    for (n=2; n<=23; n++) {
      printf("%Lu,",small_minimal_domsets_count_max (n));
    }
    printf("\n");
  }

  int num_vertices;
  for (num_vertices = 10;
       num_vertices <= 40;
       num_vertices += 1) {
    printf("gentreeg  num_vertices=%d\n", num_vertices);
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int gentreeg_argc = 0;
    char *gentreeg_argv[10];
    gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
    gentreeg_argv[gentreeg_argc++] = "-q";       /* quiet */
    gentreeg_argv[gentreeg_argc++] = num_vertices_str;
    gentreeg_argv[gentreeg_argc] = NULL;
    my_gentreeg_main(gentreeg_argc,gentreeg_argv);
  }

  return 0;
}
